@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009, 2011, 2012 CNRS.  All rights reserved.
- * Copyright © 2009-2017 Inria.  All rights reserved.
+ * Copyright © 2009-2018 Inria.  All rights reserved.
  * Copyright © 2009, 2011, 2012, 2015 Université Bordeaux.  All rights reserved.
  * Copyright © 2009 Cisco Systems, Inc.  All rights reserved.
  * $COPYRIGHT$
@@ -59,6 +59,10 @@
    */
 #define HAVE_DECL_FABSF 1
 
+/* Define to 1 if you have the declaration of `modff', and to 0 if you don't.
+   */
+#define HAVE_DECL_MODFF 1
+
 /* Define to 1 if you have the declaration of `HW_NCPU', and to 0 if you
    don't. */
 /* #undef HAVE_DECL_HW_NCPU */
@@ -81,7 +85,7 @@
 
 /* Define to 1 if you have the declaration of `strcasecmp', and to 0 if you
    don't. */
-#define HAVE_DECL_STRCASECMP 0
+/* #undef HWLOC_HAVE_DECL_STRCASECMP */
 
 /* Define to 1 if you have the declaration of `snprintf', and to 0 if you
    don't. */
@@ -661,7 +665,7 @@
 
 
 /* Version number of package */
-#define HWLOC_VERSION "1.11.9"
+#define HWLOC_VERSION "1.11.12"
 #define VERSION HWLOC_VERSION
 
 /* Define to 1 if the X Window System is missing or not being used. */
