@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2015 Inria.  All rights reserved.
+ * Copyright © 2009-2017 Inria.  All rights reserved.
  * Copyright © 2009-2011 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -317,6 +317,8 @@ int hwloc_bitmap_asprintf(char ** strp, const struct hwloc_bitmap_s * __hwloc_re
 
   len = hwloc_bitmap_snprintf(NULL, 0, set);
   buf = malloc(len+1);
+  if (!buf)
+    return -1;
   *strp = buf;
   return hwloc_bitmap_snprintf(buf, len+1, set);
 }
@@ -448,6 +450,8 @@ int hwloc_bitmap_list_asprintf(char ** strp, const struct hwloc_bitmap_s * __hwl
 
   len = hwloc_bitmap_list_snprintf(NULL, 0, set);
   buf = malloc(len+1);
+  if (!buf)
+    return -1;
   *strp = buf;
   return hwloc_bitmap_list_snprintf(buf, len+1, set);
 }
@@ -463,7 +467,7 @@ int hwloc_bitmap_list_sscanf(struct hwloc_bitmap_s *set, const char * __hwloc_re
   while (*current != '\0') {
 
     /* ignore empty ranges */
-    while (*current == ',')
+    while (*current == ',' || *current == ' ')
       current++;
 
     val = strtoul(current, &next, 0);
@@ -487,7 +491,7 @@ int hwloc_bitmap_list_sscanf(struct hwloc_bitmap_s *set, const char * __hwloc_re
 	begin = val;
       }
 
-    } else if (*next == ',' || *next == '\0') {
+    } else if (*next == ',' || *next == ' ' || *next == '\0') {
       /* single digit */
       hwloc_bitmap_set(set, val);
     }
@@ -587,6 +591,8 @@ int hwloc_bitmap_taskset_asprintf(char ** strp, const struct hwloc_bitmap_s * __
 
   len = hwloc_bitmap_taskset_snprintf(NULL, 0, set);
   buf = malloc(len+1);
+  if (!buf)
+    return -1;
   *strp = buf;
   return hwloc_bitmap_taskset_snprintf(buf, len+1, set);
 }
@@ -598,7 +604,6 @@ int hwloc_bitmap_taskset_sscanf(struct hwloc_bitmap_s *set, const char * __hwloc
   int count;
   int infinite = 0;
 
-  current = string;
   if (!strncmp("0xf...f", current, 7)) {
     /* infinite bitmap */
     infinite = 1;
@@ -771,31 +776,45 @@ void hwloc_bitmap_set_range(struct hwloc_bitmap_s * set, unsigned begincpu, int
 
 	HWLOC__BITMAP_CHECK(set);
 
-	if (_endcpu == -1) {
-		set->infinite = 1;
-		/* keep endcpu == -1 since this unsigned is actually larger than anything else */
-	}
-
-	if (set->infinite) {
-		/* truncate the range according to the infinite part of the bitmap */
-		if (endcpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
-			endcpu = set->ulongs_count * HWLOC_BITS_PER_LONG - 1;
-		if (begincpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
-			return;
-	}
 	if (endcpu < begincpu)
 		return;
-	hwloc_bitmap_realloc_by_cpu_index(set, endcpu);
+	if (set->infinite && begincpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
+		/* setting only in the already-set infinite part, nothing to do */
+		return;
 
-	beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
-	endset = HWLOC_SUBBITMAP_INDEX(endcpu);
-	for(i=beginset+1; i<endset; i++)
-		set->ulongs[i] = HWLOC_SUBBITMAP_FULL;
-	if (beginset == endset) {
-		set->ulongs[beginset] |= HWLOC_SUBBITMAP_ULBIT_FROMTO(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu), HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
-	} else {
+	if (_endcpu == -1) {
+		/* infinite range */
+
+		/* make sure we can play with the ulong that contains begincpu */
+		hwloc_bitmap_realloc_by_cpu_index(set, begincpu);
+		/* update the ulong that contains begincpu */
+		beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
 		set->ulongs[beginset] |= HWLOC_SUBBITMAP_ULBIT_FROM(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu));
-		set->ulongs[endset] |= HWLOC_SUBBITMAP_ULBIT_TO(HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		/* set ulongs after begincpu if any already allocated */
+		for(i=beginset+1; i<set->ulongs_count; i++)
+			set->ulongs[i] = HWLOC_SUBBITMAP_FULL;
+		/* mark the infinity as set */
+		set->infinite = 1;
+	} else {
+		/* finite range */
+
+		/* ignore the part of the range that overlaps with the already-set infinite part */
+		if (set->infinite && endcpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
+			endcpu = set->ulongs_count * HWLOC_BITS_PER_LONG - 1;
+		/* make sure we can play with the ulongs that contain begincpu and endcpu */
+		hwloc_bitmap_realloc_by_cpu_index(set, endcpu);
+		/* update first and last ulongs */
+		beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
+		endset = HWLOC_SUBBITMAP_INDEX(endcpu);
+		if (beginset == endset) {
+			set->ulongs[beginset] |= HWLOC_SUBBITMAP_ULBIT_FROMTO(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu), HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		} else {
+			set->ulongs[beginset] |= HWLOC_SUBBITMAP_ULBIT_FROM(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu));
+			set->ulongs[endset] |= HWLOC_SUBBITMAP_ULBIT_TO(HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		}
+		/* set ulongs in the middle of the range */
+		for(i=beginset+1; i<endset; i++)
+			set->ulongs[i] = HWLOC_SUBBITMAP_FULL;
 	}
 }
 
@@ -829,31 +848,46 @@ void hwloc_bitmap_clr_range(struct hwloc_bitmap_s * set, unsigned begincpu, int
 
 	HWLOC__BITMAP_CHECK(set);
 
-	if (_endcpu == -1) {
-		set->infinite = 0;
-		/* keep endcpu == -1 since this unsigned is actually larger than anything else */
-	}
-
-	if (!set->infinite) {
-		/* truncate the range according to the infinitely-unset part of the bitmap */
-		if (endcpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
-			endcpu = set->ulongs_count * HWLOC_BITS_PER_LONG - 1;
-		if (begincpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
-			return;
-	}
 	if (endcpu < begincpu)
 		return;
-	hwloc_bitmap_realloc_by_cpu_index(set, endcpu);
 
-	beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
-	endset = HWLOC_SUBBITMAP_INDEX(endcpu);
-	for(i=beginset+1; i<endset; i++)
-		set->ulongs[i] = HWLOC_SUBBITMAP_ZERO;
-	if (beginset == endset) {
-		set->ulongs[beginset] &= ~HWLOC_SUBBITMAP_ULBIT_FROMTO(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu), HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
-	} else {
+	if (!set->infinite && begincpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
+		/* clearing only in the already-unset infinite part, nothing to do */
+		return;
+
+	if (_endcpu == -1) {
+		/* infinite range */
+
+		/* make sure we can play with the ulong that contains begincpu */
+		hwloc_bitmap_realloc_by_cpu_index(set, begincpu);
+		/* update the ulong that contains begincpu */
+		beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
 		set->ulongs[beginset] &= ~HWLOC_SUBBITMAP_ULBIT_FROM(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu));
-		set->ulongs[endset] &= ~HWLOC_SUBBITMAP_ULBIT_TO(HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		/* clear ulong after begincpu if any already allocated */
+		for(i=beginset+1; i<set->ulongs_count; i++)
+			set->ulongs[i] = HWLOC_SUBBITMAP_ZERO;
+		/* mark the infinity as unset */
+		set->infinite = 0;
+	} else {
+		/* finite range */
+
+		/* ignore the part of the range that overlaps with the already-unset infinite part */
+		if (!set->infinite && endcpu >= set->ulongs_count * HWLOC_BITS_PER_LONG)
+			endcpu = set->ulongs_count * HWLOC_BITS_PER_LONG - 1;
+		/* make sure we can play with the ulongs that contain begincpu and endcpu */
+		hwloc_bitmap_realloc_by_cpu_index(set, endcpu);
+		/* update first and last ulongs */
+		beginset = HWLOC_SUBBITMAP_INDEX(begincpu);
+		endset = HWLOC_SUBBITMAP_INDEX(endcpu);
+		if (beginset == endset) {
+			set->ulongs[beginset] &= ~HWLOC_SUBBITMAP_ULBIT_FROMTO(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu), HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		} else {
+			set->ulongs[beginset] &= ~HWLOC_SUBBITMAP_ULBIT_FROM(HWLOC_SUBBITMAP_CPU_ULBIT(begincpu));
+			set->ulongs[endset] &= ~HWLOC_SUBBITMAP_ULBIT_TO(HWLOC_SUBBITMAP_CPU_ULBIT(endcpu));
+		}
+		/* clear ulongs in the middle of the range */
+		for(i=beginset+1; i<endset; i++)
+			set->ulongs[i] = HWLOC_SUBBITMAP_ZERO;
 	}
 }
 
