@@ -1,7 +1,7 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
- * Copyright © 2009-2011, 2013 Université Bordeaux 1
+ * Copyright © 2009-2016 Inria.  All rights reserved.
+ * Copyright © 2009-2011, 2013 Université Bordeaux
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -347,6 +347,7 @@ hwloc_aix_prepare_membind(hwloc_topology_t topology, rsethandle_t *rad, hwloc_co
   noderad = rs_alloc(RS_EMPTY);
 
   hwloc_bitmap_foreach_begin(node, nodeset)
+    /* we used MCMlevel rad number for node->os_index during lookup */
     rs_getrad(rset, noderad, MCMlevel, node, 0);
     rs_op(RS_UNION, noderad, *rad, 0, 0);
   hwloc_bitmap_foreach_end();
@@ -403,7 +404,7 @@ hwloc_aix_get_sth_membind(hwloc_topology_t topology, rstype_t what, rsid_t who,
   int res = -1;
   int depth, n, i;
 
-  depth = hwloc_get_type_depth(topology, HWLOC_OBJ_NODE);
+  depth = hwloc_get_type_depth(topology, HWLOC_OBJ_NUMANODE);
   if (depth < 0) {
     errno = EXDEV;
     return -1;
@@ -432,7 +433,7 @@ hwloc_aix_get_sth_membind(hwloc_topology_t topology, rstype_t what, rsid_t who,
 
   hwloc_bitmap_free(hwloc_set);
 
-  *policy = HWLOC_MEMBIND_DEFAULT;
+  *policy = HWLOC_MEMBIND_BIND;
   res = 0;
 
 out:
@@ -584,7 +585,7 @@ hwloc_aix_alloc_membind(hwloc_topology_t topology, size_t len, hwloc_const_nodes
   ret = ra_mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0, R_RSET, rsid, aix_policy);
 
   rs_free(rsid.at_rset);
-  return ret;
+  return ret == (void*)-1 ? NULL : ret;
 }
 #endif /* P_DEFAULT */
 
@@ -608,6 +609,9 @@ look_rset(int sdl, hwloc_obj_type_t type, struct hwloc_topology *topology, int l
   }
 
   for (i = 0; i < nbnodes; i++) {
+    hwloc_bitmap_t cpuset;
+    unsigned os_index = (unsigned) -1; /* no os_index except for PU and NUMANODE below */
+
     if (rs_getrad(rset, rad, sdl, i, 0)) {
       fprintf(stderr,"rs_getrad(%d) failed: %s\n", i, strerror(errno));
       continue;
@@ -615,18 +619,30 @@ look_rset(int sdl, hwloc_obj_type_t type, struct hwloc_topology *topology, int l
     if (!rs_getinfo(rad, R_NUMPROCS, 0))
       continue;
 
-    /* It seems logical processors are numbered from 1 here, while the
-     * bindprocessor functions numbers them from 0... */
-    obj = hwloc_alloc_setup_object(type, i - (type == HWLOC_OBJ_PU));
-    obj->cpuset = hwloc_bitmap_alloc();
-    obj->os_level = sdl;
     maxcpus = rs_getinfo(rad, R_MAXPROCS, 0);
+    cpuset = hwloc_bitmap_alloc();
     for (j = 0; j < maxcpus; j++) {
       if (rs_op(RS_TESTRESOURCE, rad, NULL, R_PROCS, j))
-	hwloc_bitmap_set(obj->cpuset, j);
+	hwloc_bitmap_set(cpuset, j);
     }
+
+    if (type == HWLOC_OBJ_PU) {
+      os_index = hwloc_bitmap_first(cpuset);
+      hwloc_debug("Found PU #%u inside node %d for sdl %d\n", os_index, i, sdl);
+      assert(hwloc_bitmap_weight(cpuset) == 1);
+    } else if (type == HWLOC_OBJ_NUMANODE) {
+      /* NUMA node os_index isn't used for binding, just use the rad number to get unique values.
+       * Note that we'll use that fact in hwloc_aix_prepare_membind(). */
+      os_index = i;
+      hwloc_debug("Using os_index #%u for NUMA node inside node %d for sdl %d\n", os_index, i, sdl);
+    }
+
+    obj = hwloc_alloc_setup_object(type, os_index);
+    obj->cpuset = cpuset;
+    obj->os_level = sdl;
+
     switch(type) {
-      case HWLOC_OBJ_NODE:
+      case HWLOC_OBJ_NUMANODE:
 	obj->nodeset = hwloc_bitmap_alloc();
 	hwloc_bitmap_set(obj->nodeset, i);
 	obj->memory.local_memory = 0; /* TODO: odd, rs_getinfo(rad, R_MEMSIZE, 0) << 10 returns the total memory ... */
@@ -634,7 +650,7 @@ look_rset(int sdl, hwloc_obj_type_t type, struct hwloc_topology *topology, int l
 	obj->memory.page_types = malloc(2*sizeof(*obj->memory.page_types));
 	memset(obj->memory.page_types, 0, 2*sizeof(*obj->memory.page_types));
 	obj->memory.page_types[0].size = hwloc_getpagesize();
-#ifdef HAVE__SC_LARGE_PAGESIZE
+#if HAVE_DECL__SC_LARGE_PAGESIZE
 	obj->memory.page_types[1].size = sysconf(_SC_LARGE_PAGESIZE);
 #endif
 	/* TODO: obj->memory.page_types[1].count = rs_getinfo(rset, R_LGPGFREE, 0) / hugepagesize */
@@ -736,7 +752,7 @@ hwloc_look_aix(struct hwloc_backend *backend)
       if (i == rs_getinfo(NULL, R_MCMSDL, 0))
 	{
 	  hwloc_debug("looking AIX node sdl %d\n", i);
-	  look_rset(i, HWLOC_OBJ_NODE, topology, i);
+	  look_rset(i, HWLOC_OBJ_NUMANODE, topology, i);
 	  known = 1;
 	}
 #      ifdef R_L2CSDL
