@@ -1,5 +1,5 @@
 /*
- * Copyright © 2010-2015 Inria.  All rights reserved.
+ * Copyright © 2010-2018 Inria.  All rights reserved.
  * Copyright © 2010-2013 Université Bordeaux
  * Copyright © 2010-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -22,10 +22,15 @@
 
 #include <private/cpuid-x86.h>
 
+#ifdef HAVE_VALGRIND_VALGRIND_H
+#include <valgrind/valgrind.h>
+#endif
+
 struct hwloc_x86_backend_data_s {
   unsigned nbprocs;
   hwloc_bitmap_t apicid_set;
   int apicid_unique;
+  int is_knl;
 };
 
 #define has_topoext(features) ((features)[6] & (1 << 22))
@@ -35,9 +40,11 @@ struct cacheinfo {
   unsigned type;
   unsigned level;
   unsigned nbthreads_sharing;
+  unsigned cacheid;
 
   unsigned linesize;
   unsigned linepart;
+  int inclusive;
   int ways;
   unsigned sets;
   unsigned long size;
@@ -69,12 +76,14 @@ struct procinfo {
 enum cpuid_type {
   intel,
   amd,
+  zhaoxin,
+  hygon,
   unknown
 };
 
 static void fill_amd_cache(struct procinfo *infos, unsigned level, int type, unsigned cpuid)
 {
-  struct cacheinfo *cache;
+  struct cacheinfo *cache, *tmpcaches;
   unsigned cachenum;
   unsigned long size = 0;
 
@@ -87,8 +96,13 @@ static void fill_amd_cache(struct procinfo *infos, unsigned level, int type, uns
   if (!size)
     return;
 
+  tmpcaches = realloc(infos->cache, (infos->numcaches+1)*sizeof(*infos->cache));
+  if (!tmpcaches)
+    /* failed to allocated, ignore that cache */
+    return;
+  infos->cache = tmpcaches;
   cachenum = infos->numcaches++;
-  infos->cache = realloc(infos->cache, infos->numcaches*sizeof(*infos->cache));
+
   cache = &infos->cache[cachenum];
 
   cache->type = type;
@@ -99,6 +113,8 @@ static void fill_amd_cache(struct procinfo *infos, unsigned level, int type, uns
     cache->nbthreads_sharing = infos->max_log_proc;
   cache->linesize = cpuid & 0xff;
   cache->linepart = 0;
+  cache->inclusive = 0; /* old AMD (K8-K10) supposed to have exclusive caches */
+
   if (level == 1) {
     cache->ways = (cpuid >> 16) & 0xff;
     if (cache->ways == 0xff)
@@ -112,21 +128,7 @@ static void fill_amd_cache(struct procinfo *infos, unsigned level, int type, uns
   cache->size = size;
   cache->sets = 0;
 
-  if (infos->cpufamilynumber== 0x10 && infos->cpumodelnumber == 0x9
-      && level == 3
-      && (cache->ways == -1 || (cache->ways % 2 == 0)) && cache->nbthreads_sharing >= 8) {
-    /* Fix AMD family 0x10 model 0x9 (Magny-Cours) with 8 or 12 cores.
-     * The L3 (and its associativity) is actually split into two halves).
-     */
-    if (cache->nbthreads_sharing == 16)
-      cache->nbthreads_sharing = 12; /* nbthreads_sharing is a power of 2 but the processor actually has 8 or 12 cores */
-    cache->nbthreads_sharing /= 2;
-    cache->size /= 2;
-    if (cache->ways != -1)
-      cache->ways /= 2;
-  }
-
-  hwloc_debug("cache L%u t%u linesize %u ways %u size %luKB\n", cache->level, cache->nbthreads_sharing, cache->linesize, cache->ways, cache->size >> 10);
+  hwloc_debug("cache L%u t%u linesize %u ways %d size %luKB\n", cache->level, cache->nbthreads_sharing, cache->linesize, cache->ways, cache->size >> 10);
 }
 
 /* Fetch information from the processor itself thanks to cpuid and store it in
@@ -170,19 +172,24 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
   _extendedmodel  = (eax>>16) & 0xf;
   _family         = (eax>>8) & 0xf;
   _extendedfamily = (eax>>20) & 0xff;
-  if ((cpuid_type == intel || cpuid_type == amd) && _family == 0xf) {
+  if ((cpuid_type == intel || cpuid_type == amd || cpuid_type == hygon) && _family == 0xf) {
     infos->cpufamilynumber = _family + _extendedfamily;
   } else {
     infos->cpufamilynumber = _family;
   }
   if ((cpuid_type == intel && (_family == 0x6 || _family == 0xf))
-      || (cpuid_type == amd && _family == 0xf)) {
+      || ((cpuid_type == amd || cpuid_type == hygon) && _family == 0xf)
+      || (cpuid_type == zhaoxin && (_family == 0x6 || _family == 0x7))) {
     infos->cpumodelnumber = _model + (_extendedmodel << 4);
   } else {
     infos->cpumodelnumber = _model;
   }
   infos->cpustepping = eax & 0xf;
 
+  if (cpuid_type == intel && infos->cpufamilynumber == 0x6 &&
+      (infos->cpumodelnumber == 0x57 || infos->cpumodelnumber == 0x85))
+    data->is_knl = 1; /* KNM is the same as KNL */
+
   /* Get cpu vendor string from cpuid 0x00 */
   memset(regs, 0, sizeof(regs));
   regs[0] = 0;
@@ -208,7 +215,7 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
   /* Get core/thread information from cpuid 0x80000008
    * (not supported on Intel)
    */
-  if (cpuid_type != intel && highest_ext_cpuid >= 0x80000008) {
+  if (cpuid_type != intel && cpuid_type != zhaoxin && highest_ext_cpuid >= 0x80000008) {
     unsigned coreidsize;
     eax = 0x80000008;
     hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
@@ -241,20 +248,42 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
    * and cache information from cpuid 0x8000001d
    * (AMD topology extension)
    */
-  if (cpuid_type != intel && has_topoext(features)) {
-    unsigned apic_id, node_id, nodes_per_proc, unit_id, cores_per_unit;
+  if (cpuid_type != intel && cpuid_type != zhaoxin && has_topoext(features)) {
+    unsigned apic_id, node_id, nodes_per_proc;
+
+    /* the code below doesn't want any other cache yet */
+    assert(!infos->numcaches);
 
     eax = 0x8000001e;
     hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
     infos->apicid = apic_id = eax;
-    infos->nodeid = node_id = ecx & 0xff;
-    nodes_per_proc = ((ecx >> 8) & 7) + 1;
-    if (nodes_per_proc > 2) {
-      hwloc_debug("warning: undefined value %d, assuming it means %d\n", nodes_per_proc, nodes_per_proc);
+
+    if (infos->cpufamilynumber == 0x16) {
+      /* ecx is reserved */
+      node_id = 0;
+      nodes_per_proc = 1;
+    } else {
+      /* AMD other families or Hygon family 18h */
+      node_id = ecx & 0xff;
+      nodes_per_proc = ((ecx >> 8) & 7) + 1;
+    }
+    infos->nodeid = node_id;
+    if ((infos->cpufamilynumber == 0x15 && nodes_per_proc > 2)
+	|| ((infos->cpufamilynumber == 0x17 || infos->cpufamilynumber == 0x18) && nodes_per_proc > 4)) {
+      hwloc_debug("warning: undefined nodes_per_proc value %u, assuming it means %u\n", nodes_per_proc, nodes_per_proc);
+    }
+
+    if (infos->cpufamilynumber <= 0x16) { /* topoext appeared in 0x15 and compute-units were only used in 0x15 and 0x16 */
+      unsigned unit_id, cores_per_unit;
+      infos->unitid = unit_id = ebx & 0xff;
+      cores_per_unit = ((ebx >> 8) & 0xff) + 1;
+      hwloc_debug("topoext %08x, %u nodes, node %u, %u cores in unit %u\n", apic_id, nodes_per_proc, node_id, cores_per_unit, unit_id);
+    } else {
+      unsigned core_id, threads_per_core;
+      infos->coreid = core_id = ebx & 0xff;
+      threads_per_core = ((ebx >> 8) & 0xff) + 1;
+      hwloc_debug("topoext %08x, %u nodes, node %u, %u threads in core %u\n", apic_id, nodes_per_proc, node_id, threads_per_core, core_id);
     }
-    infos->unitid = unit_id = ebx & 0xff;
-    cores_per_unit = ((ebx >> 8) & 3) + 1;
-    hwloc_debug("x2APIC %08x, %d nodes, node %d, %d cores in unit %d\n", apic_id, nodes_per_proc, node_id, cores_per_unit, unit_id);
 
     for (cachenum = 0; ; cachenum++) {
       unsigned type;
@@ -268,8 +297,8 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
     }
 
     cache = infos->cache = malloc(infos->numcaches * sizeof(*infos->cache));
-
-    for (cachenum = 0; ; cachenum++) {
+    if (cache) {
+     for (cachenum = 0; ; cachenum++) {
       unsigned long linesize, linepart, ways, sets;
       unsigned type;
       eax = 0x8000001d;
@@ -297,23 +326,27 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
 	cache->ways = ways;
       cache->sets = sets = ecx + 1;
       cache->size = linesize * linepart * ways * sets;
+      cache->inclusive = edx & 0x2;
 
-      hwloc_debug("cache %u type %u L%u t%u c%u linesize %lu linepart %lu ways %lu sets %lu, size %uKB\n", cachenum, cache->type, cache->level, cache->nbthreads_sharing, infos->max_nbcores, linesize, linepart, ways, sets, cache->size >> 10);
+      hwloc_debug("cache %u type %u L%u t%u c%u linesize %lu linepart %lu ways %lu sets %lu, size %luKB\n", cachenum, cache->type, cache->level, cache->nbthreads_sharing, infos->max_nbcores, linesize, linepart, ways, sets, cache->size >> 10);
 
       cache++;
+     }
+    } else {
+     infos->numcaches = 0;
     }
   } else {
     /* If there's no topoext,
      * get cache information from cpuid 0x80000005 and 0x80000006
      * (not supported on Intel)
      */
-    if (cpuid_type != intel && highest_ext_cpuid >= 0x80000005) {
+    if (cpuid_type != intel && cpuid_type != zhaoxin && highest_ext_cpuid >= 0x80000005) {
       eax = 0x80000005;
       hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
       fill_amd_cache(infos, 1, 1, ecx); /* L1d */
       fill_amd_cache(infos, 1, 2, edx); /* L1i */
     }
-    if (cpuid_type != intel && highest_ext_cpuid >= 0x80000006) {
+    if (cpuid_type != intel && cpuid_type != zhaoxin && highest_ext_cpuid >= 0x80000006) {
       eax = 0x80000006;
       hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
       if (ecx & 0xf000)
@@ -330,7 +363,11 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
   /* Get thread/core + cache information from cpuid 0x04
    * (not supported on AMD)
    */
-  if (cpuid_type != amd && highest_cpuid >= 0x04) {
+  if ((cpuid_type != amd && cpuid_type != hygon) && highest_cpuid >= 0x04) {
+    unsigned level;
+    struct cacheinfo *tmpcaches;
+    unsigned oldnumcaches = infos->numcaches; /* in case we got caches above */
+
     for (cachenum = 0; ; cachenum++) {
       unsigned type;
       eax = 0x04;
@@ -343,6 +380,10 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
 
       if (type == 0)
 	break;
+      level = (eax >> 5) & 0x7;
+      if (data->is_knl && level == 3)
+	/* KNL reports wrong L3 information (size always 0, cpuset always the entire machine, ignore it */
+	break;
       infos->numcaches++;
 
       if (!cachenum) {
@@ -356,9 +397,14 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
       }
     }
 
-    cache = infos->cache = malloc(infos->numcaches * sizeof(*infos->cache));
+    tmpcaches = realloc(infos->cache, infos->numcaches * sizeof(*infos->cache));
+    if (!tmpcaches) {
+     infos->numcaches = oldnumcaches;
+    } else {
+     infos->cache = tmpcaches;
+     cache = &infos->cache[oldnumcaches];
 
-    for (cachenum = 0; ; cachenum++) {
+     for (cachenum = 0; ; cachenum++) {
       unsigned long linesize, linepart, ways, sets;
       unsigned type;
       eax = 0x04;
@@ -369,9 +415,13 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
 
       if (type == 0)
 	break;
+      level = (eax >> 5) & 0x7;
+      if (data->is_knl && level == 3)
+	/* KNL reports wrong L3 information (size always 0, cpuset always the entire machine, ignore it */
+	break;
 
       cache->type = type;
-      cache->level = (eax >> 5) & 0x7;
+      cache->level = level;
       cache->nbthreads_sharing = ((eax >> 14) & 0xfff) + 1;
 
       cache->linesize = linesize = (ebx & 0xfff) + 1;
@@ -384,17 +434,19 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
         cache->ways = ways;
       cache->sets = sets = ecx + 1;
       cache->size = linesize * linepart * ways * sets;
+      cache->inclusive = edx & 0x2;
 
-      hwloc_debug("cache %u type %u L%u t%u c%u linesize %lu linepart %lu ways %lu sets %lu, size %uKB\n", cachenum, cache->type, cache->level, cache->nbthreads_sharing, infos->max_nbcores, linesize, linepart, ways, sets, cache->size >> 10);
+      hwloc_debug("cache %u type %u L%u t%u c%u linesize %lu linepart %lu ways %lu sets %lu, size %luKB\n", cachenum, cache->type, cache->level, cache->nbthreads_sharing, infos->max_nbcores, linesize, linepart, ways, sets, cache->size >> 10);
 
       cache++;
+     }
     }
   }
 
   /* Get package/core/thread information from cpuid 0x0b
    * (Intel x2APIC)
    */
-  if (cpuid_type == intel && has_x2apic(features)) {
+  if ((cpuid_type == intel || cpuid_type == zhaoxin) && highest_cpuid >= 0x0b && has_x2apic(features)) {
     unsigned level, apic_nextshift, apic_number, apic_type, apic_id = 0, apic_shift = 0, id;
     for (level = 0; ; level++) {
       ecx = level;
@@ -404,9 +456,10 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
         break;
     }
     if (level) {
-      infos->levels = level;
       infos->otherids = malloc(level * sizeof(*infos->otherids));
-      for (level = 0; ; level++) {
+      if (infos->otherids) {
+       infos->levels = level;
+       for (level = 0; ; level++) {
 	ecx = level;
 	eax = 0x0b;
 	hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
@@ -417,7 +470,7 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
 	apic_type = (ecx & 0xff00) >> 8;
 	apic_id = edx;
 	id = (apic_id >> apic_shift) & ((1 << (apic_nextshift - apic_shift)) - 1);
-	hwloc_debug("x2APIC %08x %d: nextshift %d num %2d type %d id %2d\n", apic_id, level, apic_nextshift, apic_number, apic_type, id);
+	hwloc_debug("x2APIC %08x %u: nextshift %u num %2u type %u id %2u\n", apic_id, level, apic_nextshift, apic_number, apic_type, id);
 	infos->apicid = apic_id;
 	infos->otherids[level] = UINT_MAX;
 	switch (apic_type) {
@@ -428,7 +481,7 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
 	  infos->coreid = id;
 	  break;
 	default:
-	  hwloc_debug("x2APIC %d: unknown type %d\n", level, apic_type);
+	  hwloc_debug("x2APIC %u: unknown type %u\n", level, apic_type);
 	  infos->otherids[level] = apic_id >> apic_shift;
 	  break;
 	}
@@ -436,8 +489,68 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
       }
       infos->apicid = apic_id;
       infos->packageid = apic_id >> apic_shift;
-      hwloc_debug("x2APIC remainder: %d\n", infos->packageid);
+      hwloc_debug("x2APIC remainder: %u\n", infos->packageid);
       hwloc_debug("this is thread %u of core %u\n", infos->threadid, infos->coreid);
+     }
+    }
+  }
+
+  /* Now that we have all info, compute cacheids and apply quirks */
+  for (cachenum = 0; cachenum < infos->numcaches; cachenum++) {
+    cache = &infos->cache[cachenum];
+
+    /* default cacheid value */
+    cache->cacheid = infos->apicid / cache->nbthreads_sharing;
+
+    if (cpuid_type == amd) {
+      /* AMD quirks */
+      if (infos->cpufamilynumber == 0x17
+	  && cache->level == 3 && cache->nbthreads_sharing == 6) {
+	/* AMD family 0x17 always shares L3 between 8 APIC ids,
+	 * even when only 6 APIC ids are enabled and reported in nbthreads_sharing
+	 * (on 24-core CPUs).
+	 */
+	cache->cacheid = infos->apicid / 8;
+
+      } else if (infos->cpufamilynumber== 0x10 && infos->cpumodelnumber == 0x9
+	  && cache->level == 3
+	  && (cache->ways == -1 || (cache->ways % 2 == 0)) && cache->nbthreads_sharing >= 8) {
+	/* Fix AMD family 0x10 model 0x9 (Magny-Cours) with 8 or 12 cores.
+	 * The L3 (and its associativity) is actually split into two halves).
+	 */
+	if (cache->nbthreads_sharing == 16)
+	  cache->nbthreads_sharing = 12; /* nbthreads_sharing is a power of 2 but the processor actually has 8 or 12 cores */
+	cache->nbthreads_sharing /= 2;
+	cache->size /= 2;
+	if (cache->ways != -1)
+	  cache->ways /= 2;
+	/* AMD Magny-Cours 12-cores processor reserve APIC ids as AAAAAABBBBBB....
+	 * among first L3 (A), second L3 (B), and unexisting cores (.).
+	 * On multi-socket servers, L3 in non-first sockets may have APIC id ranges
+	 * such as [16-21] that are not aligned on multiple of nbthreads_sharing (6).
+	 * That means, we can't just compare apicid/nbthreads_sharing to identify siblings.
+	 */
+	cache->cacheid = (infos->apicid % infos->max_log_proc) / cache->nbthreads_sharing /* cacheid within the package */
+	  + 2 * (infos->apicid / infos->max_log_proc); /* add 2 caches per previous package */
+
+      } else if (infos->cpufamilynumber == 0x15
+		 && (infos->cpumodelnumber == 0x1 /* Bulldozer */ || infos->cpumodelnumber == 0x2 /* Piledriver */)
+		 && cache->level == 3 && cache->nbthreads_sharing == 6) {
+	/* AMD Bulldozer and Piledriver 12-core processors have same APIC ids as Magny-Cours above,
+	 * but we can't merge the checks because the original nbthreads_sharing must be exactly 6 here.
+	 */
+	cache->cacheid = (infos->apicid % infos->max_log_proc) / cache->nbthreads_sharing /* cacheid within the package */
+	  + 2 * (infos->apicid / infos->max_log_proc); /* add 2 cache per previous package */
+      }
+    } else if (cpuid_type == hygon) {
+      if (infos->cpufamilynumber == 0x18
+	  && cache->level == 3 && cache->nbthreads_sharing == 6) {
+        /* Hygon family 0x18 always shares L3 between 8 APIC ids,
+         * even when only 6 APIC ids are enabled and reported in nbthreads_sharing
+         * (on 24-core CPUs).
+         */
+        cache->cacheid = infos->apicid / 8;
+      }
     }
   }
 
@@ -467,7 +580,7 @@ hwloc_x86_add_cpuinfos(hwloc_obj_t obj, struct procinfo *info, int nodup)
 }
 
 /* Analyse information stored in infos, and build/annotate topology levels accordingly */
-static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int fulldiscovery)
+static int summarize(struct hwloc_backend *backend, struct procinfo *infos, int fulldiscovery)
 {
   struct hwloc_topology *topology = backend->topology;
   struct hwloc_x86_backend_data_s *data = backend->private_data;
@@ -477,6 +590,8 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
   unsigned nbpackages = 0;
   int one = -1;
   unsigned next_group_depth = topology->next_group_depth;
+  int caches_added = 0;
+  hwloc_bitmap_t remaining_cpuset;
 
   for (i = 0; i < nbprocs; i++)
     if (infos[i].present) {
@@ -486,9 +601,11 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 
   if (one == -1) {
     hwloc_bitmap_free(complete_cpuset);
-    return;
+    return 0;
   }
 
+  remaining_cpuset = hwloc_bitmap_alloc();
+
   /* Ideally, when fulldiscovery=0, we could add any object that doesn't exist yet.
    * But what if the x86 and the native backends disagree because one is buggy? Which one to trust?
    * Only annotate existing objects for now.
@@ -496,18 +613,18 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 
   /* Look for packages */
   if (fulldiscovery) {
-    hwloc_bitmap_t packages_cpuset = hwloc_bitmap_dup(complete_cpuset);
     hwloc_bitmap_t package_cpuset;
     hwloc_obj_t package;
 
-    while ((i = hwloc_bitmap_first(packages_cpuset)) != (unsigned) -1) {
+    hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+    while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
       unsigned packageid = infos[i].packageid;
 
       package_cpuset = hwloc_bitmap_alloc();
       for (j = i; j < nbprocs; j++) {
         if (infos[j].packageid == packageid) {
           hwloc_bitmap_set(package_cpuset, j);
-          hwloc_bitmap_clr(packages_cpuset, j);
+          hwloc_bitmap_clr(remaining_cpuset, j);
         }
       }
       package = hwloc_alloc_setup_object(HWLOC_OBJ_PACKAGE, packageid);
@@ -520,7 +637,6 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
       hwloc_insert_object_by_cpuset(topology, package);
       nbpackages++;
     }
-    hwloc_bitmap_free(packages_cpuset);
 
   } else {
     /* Annotate packages previously-existing packages */
@@ -566,29 +682,29 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 
   /* Look for Numa nodes inside packages */
   if (fulldiscovery) {
-    hwloc_bitmap_t nodes_cpuset = hwloc_bitmap_dup(complete_cpuset);
     hwloc_bitmap_t node_cpuset;
     hwloc_obj_t node;
 
-    while ((i = hwloc_bitmap_first(nodes_cpuset)) != (unsigned) -1) {
+    hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+    while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
       unsigned packageid = infos[i].packageid;
       unsigned nodeid = infos[i].nodeid;
 
       if (nodeid == (unsigned)-1) {
-        hwloc_bitmap_clr(nodes_cpuset, i);
+        hwloc_bitmap_clr(remaining_cpuset, i);
 	continue;
       }
 
       node_cpuset = hwloc_bitmap_alloc();
       for (j = i; j < nbprocs; j++) {
 	if (infos[j].nodeid == (unsigned) -1) {
-	  hwloc_bitmap_clr(nodes_cpuset, j);
+	  hwloc_bitmap_clr(remaining_cpuset, j);
 	  continue;
 	}
 
         if (infos[j].packageid == packageid && infos[j].nodeid == nodeid) {
           hwloc_bitmap_set(node_cpuset, j);
-          hwloc_bitmap_clr(nodes_cpuset, j);
+          hwloc_bitmap_clr(remaining_cpuset, j);
         }
       }
       node = hwloc_alloc_setup_object(HWLOC_OBJ_NUMANODE, nodeid);
@@ -599,61 +715,59 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
           nodeid, node_cpuset);
       hwloc_insert_object_by_cpuset(topology, node);
     }
-    hwloc_bitmap_free(nodes_cpuset);
   }
 
-  /* Look for Compute units inside packages */
   if (fulldiscovery) {
-    hwloc_bitmap_t units_cpuset = hwloc_bitmap_dup(complete_cpuset);
     hwloc_bitmap_t unit_cpuset;
     hwloc_obj_t unit;
 
-    while ((i = hwloc_bitmap_first(units_cpuset)) != (unsigned) -1) {
+    /* Look for Compute units inside packages */
+    hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+    while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
       unsigned packageid = infos[i].packageid;
       unsigned unitid = infos[i].unitid;
 
       if (unitid == (unsigned)-1) {
-        hwloc_bitmap_clr(units_cpuset, i);
+        hwloc_bitmap_clr(remaining_cpuset, i);
 	continue;
       }
 
       unit_cpuset = hwloc_bitmap_alloc();
       for (j = i; j < nbprocs; j++) {
 	if (infos[j].unitid == (unsigned) -1) {
-	  hwloc_bitmap_clr(units_cpuset, j);
+	  hwloc_bitmap_clr(remaining_cpuset, j);
 	  continue;
 	}
 
         if (infos[j].packageid == packageid && infos[j].unitid == unitid) {
           hwloc_bitmap_set(unit_cpuset, j);
-          hwloc_bitmap_clr(units_cpuset, j);
+          hwloc_bitmap_clr(remaining_cpuset, j);
         }
       }
       unit = hwloc_alloc_setup_object(HWLOC_OBJ_GROUP, unitid);
       unit->cpuset = unit_cpuset;
+      hwloc_obj_add_info(unit, "Type", "ComputeUnit");
       hwloc_debug_1arg_bitmap("os unit %u has cpuset %s\n",
           unitid, unit_cpuset);
       hwloc_insert_object_by_cpuset(topology, unit);
     }
-    hwloc_bitmap_free(units_cpuset);
-  }
 
-  /* Look for unknown objects */
-  if (infos[one].otherids) {
+   /* Look for unknown objects */
+   if (infos[one].otherids) {
     for (level = infos[one].levels-1; level <= infos[one].levels-1; level--) {
       if (infos[one].otherids[level] != UINT_MAX) {
-	hwloc_bitmap_t unknowns_cpuset = hwloc_bitmap_dup(complete_cpuset);
 	hwloc_bitmap_t unknown_cpuset;
 	hwloc_obj_t unknown_obj;
 
-	while ((i = hwloc_bitmap_first(unknowns_cpuset)) != (unsigned) -1) {
+	hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+	while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
 	  unsigned unknownid = infos[i].otherids[level];
 
 	  unknown_cpuset = hwloc_bitmap_alloc();
 	  for (j = i; j < nbprocs; j++) {
 	    if (infos[j].otherids[level] == unknownid) {
 	      hwloc_bitmap_set(unknown_cpuset, j);
-	      hwloc_bitmap_clr(unknowns_cpuset, j);
+	      hwloc_bitmap_clr(remaining_cpuset, j);
 	    }
 	  }
 	  unknown_obj = hwloc_alloc_setup_object(HWLOC_OBJ_GROUP, unknownid);
@@ -662,40 +776,41 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 	  unknown_obj->attr->group.depth = topology->next_group_depth + level;
 	  if (next_group_depth <= topology->next_group_depth + level)
 	    next_group_depth = topology->next_group_depth + level + 1;
-	  hwloc_debug_2args_bitmap("os unknown%d %u has cpuset %s\n",
+	  hwloc_debug_2args_bitmap("os unknown%u %u has cpuset %s\n",
 	      level, unknownid, unknown_cpuset);
 	  hwloc_insert_object_by_cpuset(topology, unknown_obj);
 	}
-	hwloc_bitmap_free(unknowns_cpuset);
       }
     }
+   }
   }
 
   /* Look for cores */
   if (fulldiscovery) {
-    hwloc_bitmap_t cores_cpuset = hwloc_bitmap_dup(complete_cpuset);
     hwloc_bitmap_t core_cpuset;
     hwloc_obj_t core;
 
-    while ((i = hwloc_bitmap_first(cores_cpuset)) != (unsigned) -1) {
+    hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+    while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
       unsigned packageid = infos[i].packageid;
+      unsigned nodeid = infos[i].nodeid;
       unsigned coreid = infos[i].coreid;
 
       if (coreid == (unsigned) -1) {
-        hwloc_bitmap_clr(cores_cpuset, i);
+        hwloc_bitmap_clr(remaining_cpuset, i);
 	continue;
       }
 
       core_cpuset = hwloc_bitmap_alloc();
       for (j = i; j < nbprocs; j++) {
 	if (infos[j].coreid == (unsigned) -1) {
-	  hwloc_bitmap_clr(cores_cpuset, j);
+	  hwloc_bitmap_clr(remaining_cpuset, j);
 	  continue;
 	}
 
-        if (infos[j].packageid == packageid && infos[j].coreid == coreid) {
+        if (infos[j].packageid == packageid && infos[j].nodeid == nodeid && infos[j].coreid == coreid) {
           hwloc_bitmap_set(core_cpuset, j);
-          hwloc_bitmap_clr(cores_cpuset, j);
+          hwloc_bitmap_clr(remaining_cpuset, j);
         }
       }
       core = hwloc_alloc_setup_object(HWLOC_OBJ_CORE, coreid);
@@ -704,12 +819,10 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
           coreid, core_cpuset);
       hwloc_insert_object_by_cpuset(topology, core);
     }
-    hwloc_bitmap_free(cores_cpuset);
   }
 
   /* Look for PUs */
   if (fulldiscovery) {
-    unsigned i;
     hwloc_debug("%s", "\n\n * CPU cpusets *\n\n");
     for (i=0; i<nbprocs; i++)
       if(infos[i].present) { /* Only add present PU. We don't know if others actually exist */
@@ -728,18 +841,16 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
     for (j = 0; j < infos[i].numcaches; j++)
       if (infos[i].cache[j].level > level)
         level = infos[i].cache[j].level;
-
-  /* Look for known types */
-  if (fulldiscovery) while (level > 0) {
+  while (level > 0) {
     for (type = 1; type <= 3; type++) {
       /* Look for caches of that type at level level */
       {
-	hwloc_bitmap_t caches_cpuset = hwloc_bitmap_dup(complete_cpuset);
-	hwloc_bitmap_t cache_cpuset;
 	hwloc_obj_t cache;
 
-	while ((i = hwloc_bitmap_first(caches_cpuset)) != (unsigned) -1) {
-	  unsigned packageid = infos[i].packageid;
+	hwloc_bitmap_copy(remaining_cpuset, complete_cpuset);
+	while ((i = hwloc_bitmap_first(remaining_cpuset)) != (unsigned) -1) {
+	  hwloc_bitmap_t puset;
+	  int depth;
 
 	  for (l = 0; l < infos[i].numcaches; l++) {
 	    if (infos[i].cache[l].level == level && infos[i].cache[l].type == type)
@@ -747,22 +858,31 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 	  }
 	  if (l == infos[i].numcaches) {
 	    /* no cache Llevel of that type in i */
-	    hwloc_bitmap_clr(caches_cpuset, i);
+	    hwloc_bitmap_clr(remaining_cpuset, i);
 	    continue;
 	  }
 
-	  /* Found a matching cache, now look for others sharing it */
-	  {
-	    /* AMD Magny-Cours 12-cores processor reserve APIC ids as AAAAAABBBBBB....
-	     * among first L3 (A), second L3 (B), and unexisting cores (.).
-	     * On multi-socket servers, L3 in non-first sockets may have APIC id ranges
-	     * such as [16-21] that are not aligned on multiple of nbthreads_sharing (6).
-	     * That means, we can't just compare apicid/nbthreads_sharing to identify siblings.
-	     * Hence we use apicid%max_log_proc instead of apicid to restore the alignment.
-	     * Works because we also compare packageid to identify siblings.
-	     */
-	    unsigned cacheid = (infos[i].apicid % infos[i].max_log_proc) / infos[i].cache[l].nbthreads_sharing;
+	  puset = hwloc_bitmap_alloc();
+	  hwloc_bitmap_set(puset, i);
+	  depth = hwloc_get_cache_type_depth(topology, level,
+					     type == 1 ? HWLOC_OBJ_CACHE_DATA : type == 2 ? HWLOC_OBJ_CACHE_INSTRUCTION : HWLOC_OBJ_CACHE_UNIFIED);
+	  if (depth != HWLOC_TYPE_DEPTH_UNKNOWN)
+	    cache = hwloc_get_next_obj_covering_cpuset_by_depth(topology, puset, depth, NULL);
+	  else
+	    cache = NULL;
+	  hwloc_bitmap_free(puset);
 
+	  if (cache) {
+	    /* Found cache above that PU, annotate if no such attribute yet */
+	    if (!hwloc_obj_get_info_by_name(cache, "Inclusive"))
+	      hwloc_obj_add_info(cache, "Inclusive", infos[i].cache[l].inclusive ? "1" : "0");
+	    hwloc_bitmap_andnot(remaining_cpuset, remaining_cpuset, cache->cpuset);
+	  } else {
+	    /* Add the missing cache */
+	    hwloc_bitmap_t cache_cpuset;
+	    unsigned packageid = infos[i].packageid;
+	    unsigned cacheid = infos[i].cache[l].cacheid;
+	    /* Now look for others sharing it */
 	    cache_cpuset = hwloc_bitmap_alloc();
 	    for (j = i; j < nbprocs; j++) {
 	      unsigned l2;
@@ -772,12 +892,12 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 	      }
 	      if (l2 == infos[j].numcaches) {
 		/* no cache Llevel of that type in j */
-		hwloc_bitmap_clr(caches_cpuset, j);
+		hwloc_bitmap_clr(remaining_cpuset, j);
 		continue;
 	      }
-	      if (infos[j].packageid == packageid && (infos[j].apicid % infos[j].max_log_proc) / infos[j].cache[l2].nbthreads_sharing == cacheid) {
+	      if (infos[j].packageid == packageid && infos[j].cache[l2].cacheid == cacheid) {
 		hwloc_bitmap_set(cache_cpuset, j);
-		hwloc_bitmap_clr(caches_cpuset, j);
+		hwloc_bitmap_clr(remaining_cpuset, j);
 	      }
 	    }
 	    cache = hwloc_alloc_setup_object(HWLOC_OBJ_CACHE, cacheid);
@@ -797,25 +917,23 @@ static void summarize(struct hwloc_backend *backend, struct procinfo *infos, int
 		break;
 	    }
 	    cache->cpuset = cache_cpuset;
+	    hwloc_obj_add_info(cache, "Inclusive", infos[i].cache[l].inclusive ? "1" : "0");
 	    hwloc_debug_2args_bitmap("os L%u cache %u has cpuset %s\n",
 		level, cacheid, cache_cpuset);
 	    hwloc_insert_object_by_cpuset(topology, cache);
+	    caches_added++;
 	  }
 	}
-	hwloc_bitmap_free(caches_cpuset);
       }
     }
     level--;
   }
 
-  for (i = 0; i < nbprocs; i++) {
-    free(infos[i].cache);
-    if (infos[i].otherids)
-      free(infos[i].otherids);
-  }
-
+  hwloc_bitmap_free(remaining_cpuset);
   hwloc_bitmap_free(complete_cpuset);
   topology->next_group_depth = next_group_depth;
+
+  return fulldiscovery || caches_added;
 }
 
 static int
@@ -830,6 +948,7 @@ look_procs(struct hwloc_backend *backend, struct procinfo *infos, int fulldiscov
   hwloc_bitmap_t orig_cpuset = hwloc_bitmap_alloc();
   hwloc_bitmap_t set;
   unsigned i;
+  int ret = 0;
 
   if (get_cpubind(topology, orig_cpuset, HWLOC_CPUBIND_STRICT)) {
     hwloc_bitmap_free(orig_cpuset);
@@ -840,9 +959,9 @@ look_procs(struct hwloc_backend *backend, struct procinfo *infos, int fulldiscov
 
   for (i = 0; i < nbprocs; i++) {
     hwloc_bitmap_only(set, i);
-    hwloc_debug("binding to CPU%d\n", i);
+    hwloc_debug("binding to CPU%u\n", i);
     if (set_cpubind(topology, set, HWLOC_CPUBIND_STRICT)) {
-      hwloc_debug("could not bind to CPU%d: %s\n", i, strerror(errno));
+      hwloc_debug("could not bind to CPU%u: %s\n", i, strerror(errno));
       continue;
     }
     look_proc(backend, &infos[i], highest_cpuid, highest_ext_cpuid, features, cpuid_type);
@@ -854,8 +973,9 @@ look_procs(struct hwloc_backend *backend, struct procinfo *infos, int fulldiscov
 
   if (!data->apicid_unique)
     fulldiscovery = 0;
-  summarize(backend, infos, fulldiscovery);
-  return fulldiscovery; /* success, but objects added only if fulldiscovery */
+  else
+    ret = summarize(backend, infos, fulldiscovery);
+  return ret;
 }
 
 #if defined HWLOC_FREEBSD_SYS && defined HAVE_CPUSET_SETID
@@ -888,6 +1008,20 @@ static void hwloc_x86_os_state_restore(hwloc_x86_os_state_t *state __hwloc_attri
 #define AMD_EDX ('e' | ('n'<<8) | ('t'<<16) | ('i'<<24))
 #define AMD_ECX ('c' | ('A'<<8) | ('M'<<16) | ('D'<<24))
 
+/* HYGON "HygonGenuine" */
+#define HYGON_EBX ('H' | ('y'<<8) | ('g'<<16) | ('o'<<24))
+#define HYGON_EDX ('n' | ('G'<<8) | ('e'<<16) | ('n'<<24))
+#define HYGON_ECX ('u' | ('i'<<8) | ('n'<<16) | ('e'<<24))
+
+/* (Zhaoxin) CentaurHauls */
+#define ZX_EBX ('C' | ('e'<<8) | ('n'<<16) | ('t'<<24))
+#define ZX_EDX ('a' | ('u'<<8) | ('r'<<16) | ('H'<<24))
+#define ZX_ECX ('a' | ('u'<<8) | ('l'<<16) | ('s'<<24))
+
+#define SH_EBX (' ' | (' '<<8) | ('S'<<16) | ('h'<<24))
+#define SH_EDX ('a' | ('n'<<8) | ('g'<<16) | ('h'<<24))
+#define SH_ECX ('a' | ('i'<<8) | (' '<<16) | (' '<<24))
+
 /* fake cpubind for when nbprocs=1 and no binding support */
 static int fake_get_cpubind(hwloc_topology_t topology __hwloc_attribute_unused,
 			    hwloc_cpuset_t set __hwloc_attribute_unused,
@@ -927,12 +1061,12 @@ int hwloc_look_x86(struct hwloc_backend *backend, int fulldiscovery)
   memset(&hooks, 0, sizeof(hooks));
   support.membind = &memsupport;
   hwloc_set_native_binding_hooks(&hooks, &support);
-  if (hooks.get_thisproc_cpubind && hooks.set_thisproc_cpubind) {
-    get_cpubind = hooks.get_thisproc_cpubind;
-    set_cpubind = hooks.set_thisproc_cpubind;
-  } else if (hooks.get_thisthread_cpubind && hooks.set_thisthread_cpubind) {
+  if (hooks.get_thisthread_cpubind && hooks.set_thisthread_cpubind) {
     get_cpubind = hooks.get_thisthread_cpubind;
     set_cpubind = hooks.set_thisthread_cpubind;
+  } else if (hooks.get_thisproc_cpubind && hooks.set_thisproc_cpubind) {
+    get_cpubind = hooks.get_thisproc_cpubind;
+    set_cpubind = hooks.set_thisproc_cpubind;
   } else {
     /* we need binding support if there are multiple PUs */
     if (nbprocs > 1)
@@ -962,6 +1096,13 @@ int hwloc_look_x86(struct hwloc_backend *backend, int fulldiscovery)
     cpuid_type = intel;
   if (ebx == AMD_EBX && ecx == AMD_ECX && edx == AMD_EDX)
     cpuid_type = amd;
+  /* support for zhaoxin x86 cpu vendor id */
+  if (ebx == ZX_EBX && ecx == ZX_ECX && edx == ZX_EDX)
+    cpuid_type = zhaoxin;
+  if (ebx == SH_EBX && ecx == SH_ECX && edx == SH_EDX)
+    cpuid_type = zhaoxin;
+  else if (ebx == HYGON_EBX && ecx == HYGON_ECX && edx == HYGON_EDX)
+    cpuid_type = hygon;
 
   hwloc_debug("highest cpuid %x, cpuid type %u\n", highest_cpuid, cpuid_type);
   if (highest_cpuid < 0x01) {
@@ -981,6 +1122,7 @@ int hwloc_look_x86(struct hwloc_backend *backend, int fulldiscovery)
 
   if (highest_cpuid >= 0x7) {
     eax = 0x7;
+    ecx = 0;
     hwloc_x86_cpuid(&eax, &ebx, &ecx, &edx);
     features[9] = ebx;
   }
@@ -1004,8 +1146,7 @@ int hwloc_look_x86(struct hwloc_backend *backend, int fulldiscovery)
   if (nbprocs == 1) {
     /* only one processor, no need to bind */
     look_proc(backend, &infos[0], highest_cpuid, highest_ext_cpuid, features, cpuid_type);
-    summarize(backend, infos, fulldiscovery);
-    ret = fulldiscovery;
+    ret = summarize(backend, infos, fulldiscovery);
   }
 
 out_with_os_state:
@@ -1013,7 +1154,12 @@ out_with_os_state:
 
 out_with_infos:
   if (NULL != infos) {
-      free(infos);
+    for (i = 0; i < nbprocs; i++) {
+      free(infos[i].cache);
+      if (infos[i].otherids)
+	free(infos[i].otherids);
+    }
+    free(infos);
   }
 
 out:
@@ -1028,6 +1174,13 @@ hwloc_x86_discover(struct hwloc_backend *backend)
   int alreadypus = 0;
   int ret;
 
+#if HAVE_DECL_RUNNING_ON_VALGRIND
+  if (RUNNING_ON_VALGRIND) {
+    fprintf(stderr, "hwloc x86 backend cannot work under Valgrind, disabling.\n");
+    return 0;
+  }
+#endif
+
   data->nbprocs = hwloc_fallback_nbprocessors(topology);
 
   if (!topology->is_thissystem) {
@@ -1043,11 +1196,11 @@ hwloc_x86_discover(struct hwloc_backend *backend)
       goto fulldiscovery;
     }
 
-    /* several object types were added, we can't easily complete, just annotate a bit */
+    /* several object types were added, we can't easily complete, just do partial discovery */
     ret = hwloc_look_x86(backend, 0);
     if (ret)
       hwloc_obj_add_info(topology->levels[0][0], "Backend", "x86");
-    return 0;
+    return ret;
   } else {
     /* topology is empty, initialize it */
     hwloc_alloc_obj_cpusets(topology->levels[0][0]);
@@ -1108,6 +1261,7 @@ hwloc_x86_component_instantiate(struct hwloc_disc_component *component,
   backend->disable = hwloc_x86_backend_disable;
 
   /* default values */
+  data->is_knl = 0;
   data->apicid_set = hwloc_bitmap_alloc();
   data->apicid_unique = 1;
 
