@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
+ * Copyright © 2009-2018 Inria.  All rights reserved.
  * Copyright © 2009-2012 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -26,17 +26,24 @@ static int show_ancestors = 0;
 static hwloc_obj_type_t show_ancestor_type = (hwloc_obj_type_t) -1;
 static int show_ancestor_attrdepth = -1;
 static hwloc_obj_cache_type_t show_ancestor_attrcachetype = (hwloc_obj_cache_type_t) -1;
+static int show_children = 0;
+static int show_descendants_depth = HWLOC_TYPE_DEPTH_UNKNOWN;
 static int show_index_prefix = 0;
-static int current_obj;
+static unsigned current_obj;
 
 void usage(const char *name, FILE *where)
 {
   fprintf (where, "Usage: %s [ options ] [ locations ]\n", name);
   fprintf (where, "\nOutput options:\n");
+  fprintf (where, "  --objects             Report information about specific objects\n");
+  fprintf (where, "  --topology            Report information the topology\n");
+  fprintf (where, "  --support             Report information about supported features\n");
   fprintf (where, "  -v --verbose          Include additional details\n");
   fprintf (where, "  -s --silent           Reduce the amount of details to show\n");
   fprintf (where, "  --ancestors           Display the chain of ancestor objects up to the root\n");
   fprintf (where, "  --ancestor <type>     Only display the ancestor of the given type\n");
+  fprintf (where, "  --children            Display all children\n");
+  fprintf (where, "  --descendants <type>  Only display descendants of the given type\n");
   fprintf (where, "  -n                    Prefix each line with the index of the considered object\n");
   fprintf (where, "Object filtering options:\n");
   fprintf (where, "  --restrict <cpuset>   Restrict the topology to processors listed in <cpuset>\n");
@@ -72,7 +79,7 @@ hwloc_info_show_obj(hwloc_obj_t obj, const char *type, const char *prefix, int v
   if (obj->name)
     printf("%s name = %s\n", prefix, obj->name);
   if (obj->depth != (unsigned) -1)
-    printf("%s depth = %u\n", prefix, obj->depth);
+    printf("%s depth = %d\n", prefix, (int) obj->depth); /* special levels have negative values */
   printf("%s sibling rank = %u\n", prefix, obj->sibling_rank);
   printf("%s children = %u\n", prefix, obj->arity);
   if (obj->memory.local_memory)
@@ -174,16 +181,21 @@ hwloc_info_show_obj(hwloc_obj_t obj, const char *type, const char *prefix, int v
     /* nothing to show */
     break;
   }
+
+  printf("%s symmetric subtree = %d\n", prefix, obj->symmetric_subtree);
+
   for(i=0; i<obj->infos_count; i++) {
     printf("%s info %s = %s\n", prefix, obj->infos[i].name, obj->infos[i].value);
   }
 }
 
 static void
-hwloc_calc_process_arg_info_cb(void *_data __hwloc_attribute_unused,
-			       hwloc_obj_t obj,
-			       int verbose)
+hwloc_calc_process_location_info_cb(struct hwloc_calc_location_context_s *lcontext,
+				    void *_data __hwloc_attribute_unused,
+				    hwloc_obj_t obj)
 {
+  hwloc_topology_t topology = lcontext->topology;
+  int verbose = lcontext->verbose;
   char prefix[32];
   char objs[128];
 
@@ -241,6 +253,73 @@ hwloc_calc_process_arg_info_cb(void *_data __hwloc_attribute_unused,
 next:
       parent = parent->parent;
     }
+  } else if (show_children) {
+    unsigned i = 0;
+    hwloc_obj_t child = NULL;
+    while ((child = hwloc_get_next_child(topology, obj, child)) != NULL) {
+      char childs[128];
+      if (show_index_prefix)
+	snprintf(prefix, sizeof(prefix), "%u.%u: ", current_obj, i);
+      hwloc_obj_type_snprintf(childs, sizeof(childs), child, 1);
+      if (verbose < 0)
+	printf("%s%s:%u\n", prefix, childs, child->logical_index);
+      else
+	printf("%s%s L#%u = child #%u of %s L#%u\n",
+	       prefix, childs, child->logical_index, i, objs, obj->logical_index);
+      hwloc_info_show_obj(child, childs, prefix, verbose);
+      i++;
+    }
+  } else if (show_descendants_depth != HWLOC_TYPE_DEPTH_UNKNOWN) {
+    if (show_descendants_depth >= 0) {
+      /* normal level */
+      unsigned i = 0;
+      unsigned n = hwloc_calc_get_nbobjs_inside_sets_by_depth(lcontext, obj->cpuset, obj->nodeset, show_descendants_depth);
+      for(i=0; i<n; i++) {
+	hwloc_obj_t child = hwloc_calc_get_obj_inside_sets_by_depth(lcontext, obj->cpuset, obj->nodeset, show_descendants_depth, i);
+	char childs[128];
+	if (show_index_prefix)
+	  snprintf(prefix, sizeof(prefix), "%u.%u: ", current_obj, i);
+	hwloc_obj_type_snprintf(childs, sizeof(childs), child, 1);
+	if (verbose < 0)
+	  printf("%s%s:%u\n", prefix, childs, child->logical_index);
+	else
+	  printf("%s%s L#%u = descendant #%u of %s L#%u\n",
+		 prefix, childs, child->logical_index, i, objs, obj->logical_index);
+	hwloc_info_show_obj(child, childs, prefix, verbose);
+      }
+    } else {
+      /* custom level */
+      unsigned i = 0;
+      hwloc_obj_t child = NULL;
+      while ((child = hwloc_get_next_obj_by_depth(topology, show_descendants_depth, child)) != NULL) {
+	char childs[128];
+	hwloc_obj_t parent = child->parent;
+	if (obj->cpuset) {
+	  while (parent && !parent->cpuset)
+	    parent = parent->parent;
+	  if (!parent)
+	    continue;
+	  if (!hwloc_bitmap_isincluded(parent->cpuset, obj->cpuset)
+	      || !hwloc_bitmap_isincluded(parent->nodeset, obj->nodeset))
+	    continue;
+	} else {
+	  while (parent && parent != obj)
+	    parent = parent->parent;
+	  if (!parent)
+	    continue;
+	}
+	if (show_index_prefix)
+	  snprintf(prefix, sizeof(prefix), "%u.%u: ", current_obj, i);
+	hwloc_obj_type_snprintf(childs, sizeof(childs), child, 1);
+	if (verbose < 0)
+	  printf("%s%s:%u\n", prefix, childs, child->logical_index);
+	else
+	  printf("%s%s L#%u = descendant #%u of %s L#%u\n",
+		 prefix, childs, child->logical_index, i, objs, obj->logical_index);
+	hwloc_info_show_obj(child, childs, prefix, verbose);
+	i++;
+      }
+    }
   } else {
     if (verbose < 0)
       printf("%s%s:%u\n", prefix, objs, obj->logical_index);
@@ -262,13 +341,17 @@ main (int argc, char *argv[])
   char * callname;
   char * input = NULL;
   enum hwloc_utils_input_format input_format = HWLOC_UTILS_INPUT_DEFAULT;
+  hwloc_obj_type_t show_descendants_type = (hwloc_obj_type_t) -1;
+  int show_descendants_attrdepth = -1;
+  hwloc_obj_cache_type_t show_descendants_attrcachetype = (hwloc_obj_cache_type_t) -1;
   char *restrictstring = NULL;
   size_t typelen;
   int opt;
+  enum hwloc_info_mode { HWLOC_INFO_MODE_UNKNOWN, HWLOC_INFO_MODE_TOPOLOGY, HWLOC_INFO_MODE_OBJECTS, HWLOC_INFO_MODE_SUPPORT } mode = HWLOC_INFO_MODE_UNKNOWN;
 
   /* enable verbose backends */
-  putenv("HWLOC_XML_VERBOSE=1");
-  putenv("HWLOC_SYNTHETIC_VERBOSE=1");
+  putenv((char *) "HWLOC_XML_VERBOSE=1");
+  putenv((char *) "HWLOC_SYNTHETIC_VERBOSE=1");
 
   callname = strrchr(argv[0], '/');
   if (!callname)
@@ -288,7 +371,13 @@ main (int argc, char *argv[])
   while (argc >= 1) {
     opt = 0;
     if (*argv[0] == '-') {
-      if (!strcmp (argv[0], "-v") || !strcmp (argv[0], "--verbose"))
+      if (!strcmp (argv[0], "--objects"))
+	mode = HWLOC_INFO_MODE_OBJECTS;
+      else if (!strcmp (argv[0], "--topology"))
+	mode = HWLOC_INFO_MODE_TOPOLOGY;
+      else if (!strcmp (argv[0], "--support"))
+	mode = HWLOC_INFO_MODE_SUPPORT;
+      else if (!strcmp (argv[0], "-v") || !strcmp (argv[0], "--verbose"))
         verbose_mode++;
       else if (!strcmp (argv[0], "-s") || !strcmp (argv[0], "--silent"))
         verbose_mode--;
@@ -309,7 +398,22 @@ main (int argc, char *argv[])
         if (err < 0) {
           fprintf(stderr, "unrecognized --ancestor type %s\n", argv[1]);
           usage(callname, stderr);
-          return EXIT_SUCCESS;
+          return EXIT_FAILURE;
+        }
+	opt = 1;
+      }
+      else if (!strcmp (argv[0], "--children"))
+	show_children = 1;
+      else if (!strcmp (argv[0], "--descendants")) {
+	if (argc < 2) {
+	  usage (callname, stderr);
+	  exit(EXIT_FAILURE);
+	}
+	err = hwloc_obj_type_sscanf(argv[1], &show_descendants_type, &show_descendants_attrdepth, &show_descendants_attrcachetype, sizeof(show_descendants_attrcachetype));
+        if (err < 0) {
+          fprintf(stderr, "unrecognized --descendants type %s\n", argv[1]);
+          usage(callname, stderr);
+          return EXIT_FAILURE;
         }
 	opt = 1;
       }
@@ -370,14 +474,14 @@ main (int argc, char *argv[])
   hwloc_topology_set_flags(topology, flags);
 
   if (input) {
-    err = hwloc_utils_enable_input_format(topology, input, input_format, verbose_mode, callname);
+    err = hwloc_utils_enable_input_format(topology, input, &input_format, verbose_mode, callname);
     if (err)
       return err;
   }
 
   if (pid_number > 0) {
-    pid = hwloc_pid_from_number(pid_number, 0);
-    if (hwloc_topology_set_pid(topology, pid)) {
+    if (hwloc_pid_from_number(&pid, pid_number, 0, 1 /* verbose */) < 0
+	|| hwloc_topology_set_pid(topology, pid)) {
       perror("Setting target pid");
       return EXIT_FAILURE;
     }
@@ -389,6 +493,32 @@ main (int argc, char *argv[])
 
   topodepth = hwloc_topology_get_depth(topology);
 
+  if (show_descendants_type != (hwloc_obj_type_t) -1) {
+    show_descendants_depth = hwloc_get_type_depth(topology, show_descendants_type);
+    if (show_descendants_depth == HWLOC_TYPE_DEPTH_UNKNOWN) {
+      fprintf(stderr, "unavailable --descendants type %s\n", hwloc_obj_type_string(show_descendants_type));
+      return EXIT_FAILURE;
+    }
+    if (show_descendants_depth == HWLOC_TYPE_DEPTH_MULTIPLE) {
+      if (show_descendants_type == HWLOC_OBJ_CACHE)
+	show_descendants_depth = hwloc_get_cache_type_depth(topology, show_descendants_attrdepth, show_descendants_attrcachetype);
+      else if (show_descendants_type == HWLOC_OBJ_GROUP) {
+	unsigned i;
+	for(i=0; i<hwloc_topology_get_depth(topology); i++) {
+	  hwloc_obj_t obj = hwloc_get_obj_by_depth(topology, i, 0);
+	  if (obj->type == HWLOC_OBJ_GROUP && obj->attr->group.depth == (unsigned) show_descendants_attrdepth) {
+	    show_descendants_depth = i;
+	    break;
+	  }
+        }
+      }
+    }
+    if (show_descendants_depth == HWLOC_TYPE_DEPTH_MULTIPLE) {
+      fprintf(stderr, "multiple --descendants type %s\n", hwloc_obj_type_string(show_descendants_type));
+      return EXIT_FAILURE;
+    }
+  }
+
   if (restrictstring) {
     hwloc_bitmap_t restrictset = hwloc_bitmap_alloc();
     if (!strcmp (restrictstring, "binding")) {
@@ -402,32 +532,78 @@ main (int argc, char *argv[])
     err = hwloc_topology_restrict (topology, restrictset, 0);
     if (err) {
       perror("Restricting the topology");
-      /* fallthrough */
+      /* FALLTHRU */
     }
     hwloc_bitmap_free(restrictset);
     free(restrictstring);
   }
 
-  if (!argc)
+  if (mode == HWLOC_INFO_MODE_UNKNOWN) {
+    if (argc)
+      mode = HWLOC_INFO_MODE_OBJECTS;
+    else
+      mode = HWLOC_INFO_MODE_TOPOLOGY;
+  }
+
+  if (mode == HWLOC_INFO_MODE_TOPOLOGY) {
     hwloc_lstopo_show_summary(stdout, topology);
 
-  else {
+  } else if (mode == HWLOC_INFO_MODE_SUPPORT) {
+    const struct hwloc_topology_support *support = hwloc_topology_get_support(topology);
+#define DO(x,y) printf(#x ":" #y " = %u\n", (unsigned char) support->x->y);
+    DO(discovery, pu);
+
+    DO(cpubind, set_thisproc_cpubind);
+    DO(cpubind, get_thisproc_cpubind);
+    DO(cpubind, set_proc_cpubind);
+    DO(cpubind, get_proc_cpubind);
+    DO(cpubind, set_thisthread_cpubind);
+    DO(cpubind, get_thisthread_cpubind);
+    DO(cpubind, set_thread_cpubind);
+    DO(cpubind, get_thread_cpubind);
+    DO(cpubind, get_thisproc_last_cpu_location);
+    DO(cpubind, get_proc_last_cpu_location);
+    DO(cpubind, get_thisthread_last_cpu_location);
+
+    DO(membind, set_thisproc_membind);
+    DO(membind, get_thisproc_membind);
+    DO(membind, set_proc_membind);
+    DO(membind, get_proc_membind);
+    DO(membind, set_thisthread_membind);
+    DO(membind, get_thisthread_membind);
+    DO(membind, set_area_membind);
+    DO(membind, get_area_membind);
+    DO(membind, alloc_membind);
+    DO(membind, firsttouch_membind);
+    DO(membind, bind_membind);
+    DO(membind, interleave_membind);
+    DO(membind, nexttouch_membind);
+    DO(membind, migrate_membind);
+    DO(membind, get_area_memlocation);
+
+  } else if (mode == HWLOC_INFO_MODE_OBJECTS) {
+    struct hwloc_calc_location_context_s lcontext;
+    lcontext.topology = topology;
+    lcontext.topodepth = topodepth;
+    lcontext.only_hbm = -1;
+    lcontext.logical = logical;
+    lcontext.verbose = verbose_mode;
     current_obj = 0;
     while (argc >= 1) {
       if (!strcmp(argv[0], "all") || !strcmp(argv[0], "root")) {
-	hwloc_calc_process_arg_info_cb(NULL, hwloc_get_root_obj(topology), verbose_mode);
+	hwloc_calc_process_location_info_cb(&lcontext, NULL, hwloc_get_root_obj(topology));
       } else {
 	/* try to match a type/depth followed by a special character */
 	typelen = strspn(argv[0], "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
 	if (typelen && (argv[0][typelen] == ':' || argv[0][typelen] == '=' || argv[0][typelen] == '[')) {
-	  err = hwloc_calc_process_type_arg(topology, topodepth, argv[0], typelen, logical,
-					    hwloc_calc_process_arg_info_cb, NULL,
-					    verbose_mode);
+	  err = hwloc_calc_process_location(&lcontext, argv[0], typelen,
+					    hwloc_calc_process_location_info_cb, NULL);
 	}
       }
       argc--; argv++;
     }
-  }
+
+  } else assert(0);
 
   hwloc_topology_destroy (topology);
 
