@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2017 Inria.  All rights reserved.
+ * Copyright © 2009-2018 Inria.  All rights reserved.
  * Copyright © 2009-2012 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -493,6 +493,11 @@ void output_ascii(struct lstopo_output *loutput, const char *filename)
     return;
   }
 
+  if (gridsize <= 1) { /* we divide by gridsize or gridsize/2 in the code */
+    fprintf(stderr, "ASCII backend requires gridsize > 1\n");
+    return;
+  }
+
 #ifdef HWLOC_HAVE_LIBTERMCAP
   /* If we are outputing to a tty, use colors */
   if (output == stdout && isatty(STDOUT_FILENO)) {
