@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2016 Inria.  All rights reserved.
+ * Copyright © 2009-2017 Inria.  All rights reserved.
  * Copyright © 2009 Université Bordeaux
  * See COPYING in top-level directory.
  */
@@ -38,7 +38,7 @@ rgb_to_color(int r, int g, int b)
   if (color != -1)
     return color;
 
-  fprintf(stderr, "color #%02x%02x%02x not declared\n", r, g, b);
+  fprintf(stderr, "color #%02x%02x%02x not declared\n", (unsigned) r, (unsigned) g, (unsigned) b);
   exit(EXIT_FAILURE);
 }
 
@@ -46,12 +46,21 @@ int
 declare_color(int r, int g, int b)
 {
   int color = find_color(r, g, b);
+  struct color *tmp;
 
   if (color != -1)
     return color;
 
+  tmp = realloc(colors, sizeof(*colors) * (numcolors+1));
+
+  if (!tmp) {
+    /* FIXME: return -1 on error? tell the caller to use the default? */
+    fprintf(stderr, "Failed to realloc the colors array\n");
+    exit(EXIT_FAILURE);
+  }
+
+  colors = tmp;
   color = numcolors++;
-  colors = realloc(colors, sizeof(*colors) * (numcolors));
   colors[color].r = r;
   colors[color].g = g;
   colors[color].b = b;
